/* LIBPIKA - The PIKA Library
 * Copyright (C) 1995-1997 Peter Mattis and Spencer Kimball
 *
 * pikalabelstringwidget.c
 * Copyright (C) 2023 Jehan
 *
 * This library is free software: you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library.  If not, see
 * <https://www.gnu.org/licenses/>.
 */

#include "config.h"

#include <gegl.h>
#include <gtk/gtk.h>

#include "libpikacolor/pikacolor.h"
#include "libpikamath/pikamath.h"
#include "libpikabase/pikabase.h"

#include "pikalabelstringwidget.h"


/**
 * SECTION: pikalabelstringwidget
 * @title: PikaLabelStringWidget
 * @short_description: Widget containing a label and a widget with a
 *                     string "value" property.
 *
 * This widget is a subclass of #PikaLabeled.
 **/

enum
{
  VALUE_CHANGED,
  LAST_SIGNAL
};

enum
{
  PROP_0,
  PROP_VALUE,
  PROP_WIDGET,
};

typedef struct _PikaLabelStringWidgetPrivate
{
  PikaLabeled    parent_instance;

  GtkWidget     *widget;
  gchar         *value;
} PikaLabelStringWidgetPrivate;

static void        pika_label_string_widget_constructed       (GObject       *object);
static void        pika_label_string_widget_finalize          (GObject       *object);
static void        pika_label_string_widget_set_property      (GObject       *object,
                                                               guint          property_id,
                                                               const GValue  *value,
                                                               GParamSpec    *pspec);
static void        pika_label_string_widget_get_property      (GObject       *object,
                                                               guint          property_id,
                                                               GValue        *value,
                                                               GParamSpec    *pspec);

static GtkWidget * pika_label_string_widget_populate          (PikaLabeled   *widget,
                                                               gint          *x,
                                                               gint          *y,
                                                               gint          *width,
                                                               gint          *height);

G_DEFINE_TYPE_WITH_PRIVATE (PikaLabelStringWidget, pika_label_string_widget, PIKA_TYPE_LABELED)

#define parent_class pika_label_string_widget_parent_class

static guint pika_label_string_widget_signals[LAST_SIGNAL] = { 0 };

static void
pika_label_string_widget_class_init (PikaLabelStringWidgetClass *klass)
{
  GObjectClass     *object_class  = G_OBJECT_CLASS (klass);
  PikaLabeledClass *labeled_class = PIKA_LABELED_CLASS (klass);

  pika_label_string_widget_signals[VALUE_CHANGED] =
    g_signal_new ("value-changed",
                  G_TYPE_FROM_CLASS (klass),
                  G_SIGNAL_RUN_FIRST,
                  G_STRUCT_OFFSET (PikaLabelStringWidgetClass, value_changed),
                  NULL, NULL, NULL,
                  G_TYPE_NONE, 0);

  object_class->constructed  = pika_label_string_widget_constructed;
  object_class->finalize     = pika_label_string_widget_finalize;
  object_class->set_property = pika_label_string_widget_set_property;
  object_class->get_property = pika_label_string_widget_get_property;

  labeled_class->populate    = pika_label_string_widget_populate;

  /**
   * PikaLabelStringWidget:value:
   *
   * The currently set value.
   *
   * Since: 3.0
   **/
  g_object_class_install_property (object_class, PROP_VALUE,
                                   g_param_spec_string ("value", NULL,
                                                        "Current value",
                                                        NULL,
                                                        PIKA_PARAM_READWRITE));

  /**
   * PikaLabelStringWidget:widget:
   *
   * The widget holding a string property named "value".
   *
   * Since: 3.0
   **/
  g_object_class_install_property (object_class, PROP_WIDGET,
                                   g_param_spec_object ("widget", NULL,
                                                        "String widget",
                                                        GTK_TYPE_WIDGET,
                                                        PIKA_PARAM_READWRITE |
                                                        G_PARAM_CONSTRUCT_ONLY));
}

static void
pika_label_string_widget_init (PikaLabelStringWidget *widget)
{
}

static void
pika_label_string_widget_constructed (GObject *object)
{
  PikaLabelStringWidget        *widget = PIKA_LABEL_STRING_WIDGET (object);
  PikaLabelStringWidgetPrivate *priv   = pika_label_string_widget_get_instance_private (widget);

  G_OBJECT_CLASS (parent_class)->constructed (object);

  gtk_grid_set_column_spacing (GTK_GRID (widget), 6);
  gtk_grid_set_row_spacing (GTK_GRID (widget), 6);

  /* This is important to make this object into a property widget. It
   * will allow config object to bind the "value" property of this
   * widget, and therefore be updated automatically.
   */
  g_object_bind_property (G_OBJECT (priv->widget), "value",
                          object,                  "value",
                          G_BINDING_BIDIRECTIONAL |
                          G_BINDING_SYNC_CREATE);
}

static void
pika_label_string_widget_finalize (GObject *object)
{
  PikaLabelStringWidget        *widget = PIKA_LABEL_STRING_WIDGET (object);
  PikaLabelStringWidgetPrivate *priv   = pika_label_string_widget_get_instance_private (widget);

  g_free (priv->value);

  G_OBJECT_CLASS (parent_class)->finalize (object);
}

static void
pika_label_string_widget_set_property (GObject      *object,
                                       guint         property_id,
                                       const GValue *value,
                                       GParamSpec   *pspec)
{
  PikaLabelStringWidget        *widget = PIKA_LABEL_STRING_WIDGET (object);
  PikaLabelStringWidgetPrivate *priv   = pika_label_string_widget_get_instance_private (widget);

  switch (property_id)
    {
    case PROP_VALUE:
      if (g_strcmp0 (priv->value, g_value_get_string (value)) != 0)
        {
          g_free (priv->value);
          priv->value = g_value_dup_string (value);
          g_signal_emit (object, pika_label_string_widget_signals[VALUE_CHANGED], 0);
        }
      break;
    case PROP_WIDGET:
      priv->widget = g_value_get_object (value);
      break;

    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
      break;
    }
}

static void
pika_label_string_widget_get_property (GObject    *object,
                                       guint       property_id,
                                       GValue     *value,
                                       GParamSpec *pspec)
{
  PikaLabelStringWidget        *widget = PIKA_LABEL_STRING_WIDGET (object);
  PikaLabelStringWidgetPrivate *priv   = pika_label_string_widget_get_instance_private (widget);

  switch (property_id)
    {
    case PROP_VALUE:
      g_value_set_string (value, priv->value);
      break;
    case PROP_WIDGET:
      g_value_set_object (value, priv->widget);
      break;

    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
      break;
    }
}

static GtkWidget *
pika_label_string_widget_populate (PikaLabeled *labeled,
                                   gint        *x,
                                   gint        *y,
                                   gint        *width,
                                   gint        *height)
{
  PikaLabelStringWidget        *widget = PIKA_LABEL_STRING_WIDGET (labeled);
  PikaLabelStringWidgetPrivate *priv   = pika_label_string_widget_get_instance_private (widget);

  gtk_grid_attach (GTK_GRID (widget), priv->widget, 1, 0, 1, 1);
  gtk_widget_show (priv->widget);

  return priv->widget;
}

/* Public Functions */

/**
 * pika_label_string_widget_new:
 * @text:   The text for the #GtkLabel.
 * @widget: (transfer full): The #GtkWidget to use.
 *
 * Creates a new #PikaLabelStringWidget whose "value" property is bound to
 * that of @widget (which must therefore have such a string property).
 *
 * Returns: (transfer full): The new #PikaLabelStringWidget widget.
 **/
GtkWidget *
pika_label_string_widget_new (const gchar *text,
                              GtkWidget   *widget)
{
  GtkWidget  *string_widget;
  GParamSpec *pspec;

  g_return_val_if_fail ((pspec = g_object_class_find_property (G_OBJECT_GET_CLASS (widget),
                                                               "value")) &&
                        (G_PARAM_SPEC_TYPE (pspec) == G_TYPE_PARAM_STRING ||
                         G_PARAM_SPEC_TYPE (pspec) == PIKA_TYPE_PARAM_CHOICE),
                        NULL);

  string_widget = g_object_new (PIKA_TYPE_LABEL_STRING_WIDGET,
                                "label",  text,
                                "widget", widget,
                                NULL);

  return string_widget;
}

/**
 * pika_label_string_widget_get_widget:
 * @widget: the #PikaLabelStringWidget.
 *
 * Returns: (transfer none): The new #GtkWidget packed next to the label.
 **/
GtkWidget *
pika_label_string_widget_get_widget (PikaLabelStringWidget *widget)
{
  PikaLabelStringWidgetPrivate *priv;

  g_return_val_if_fail (PIKA_IS_LABEL_STRING_WIDGET (widget), NULL);

  priv = pika_label_string_widget_get_instance_private (widget);

  return priv->widget;
}
